// This file is part of the re-motion Core Framework (www.re-motion.org)
// Copyright (C) 2005-2009 rubicon informationstechnologie gmbh, www.rubicon.eu
// 
// The re-motion Core Framework is free software; you can redistribute it 
// and/or modify it under the terms of the GNU Lesser General Public License 
// as published by the Free Software Foundation; either version 2.1 of the 
// License, or (at your option) any later version.
// 
// re-motion is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with re-motion; if not, see http://www.gnu.org/licenses.
// 
using System;
using System.Collections.Generic;
using System.Linq.Expressions;
using NUnit.Framework;
using Remotion.Linq.UnitTests.Linq.Core.Parsing;
using Remotion.Linq.UnitTests.Linq.Core.TestDomain;
using Remotion.Linq.UnitTests.Linq.SqlBackend.SqlStatementModel;
using Remotion.Linq.SqlBackend.SqlGeneration;
using Remotion.Linq.SqlBackend.SqlStatementModel;
using Remotion.Linq.SqlBackend.SqlStatementModel.Resolved;
using Rhino.Mocks;

namespace Remotion.Linq.UnitTests.Linq.SqlBackend.SqlGeneration
{
  [TestFixture]
  public class SqlGeneratingOuterSelectExpressionVisitorTest
  {
    private NamedExpression _namedExpression;
    private SqlEntityDefinitionExpression _entityExpression;
    private TestableSqlGeneratingOuterSelectExpressionVisitor _visitor;
    private ISqlGenerationStage _stageMock;
    private SqlCommandBuilder _commandBuilder;

    [SetUp]
    public void SetUp ()
    {
      _stageMock = MockRepository.GenerateStrictMock<ISqlGenerationStage> ();
      _commandBuilder = new SqlCommandBuilder ();
      _visitor = new TestableSqlGeneratingOuterSelectExpressionVisitor (_commandBuilder, _stageMock);

      _namedExpression = new NamedExpression ("test", Expression.Constant (0));
      _entityExpression = new SqlEntityDefinitionExpression (
          typeof (Cook),
          "c",
          "test",
          new SqlColumnDefinitionExpression (typeof (int), "c", "ID", true),
          new SqlColumnDefinitionExpression (typeof (int), "c", "ID", true),
          new SqlColumnDefinitionExpression (typeof (string), "c", "Name", false),
          new SqlColumnDefinitionExpression (typeof (string), "c", "FirstName", false)
          );
    }

    [Test]
    [ExpectedException (typeof (NotSupportedException), ExpectedMessage =
        "Queries selecting collections are not supported because SQL is not well-suited to returning collections.",
        MatchType = MessageMatch.Contains)]
    public void GenerateSql_Collection ()
    {
      var expression = Expression.Constant (new Cook[] { });
      SqlGeneratingOuterSelectExpressionVisitor.GenerateSql (expression, _commandBuilder, _stageMock);
    }

    [Test]
    public void GenerateSql_CreatesFullInMemoryProjectionLambda ()
    {
       SqlGeneratingOuterSelectExpressionVisitor.GenerateSql (_namedExpression, _commandBuilder, _stageMock);

      var expectedRowParameter = _commandBuilder.InMemoryProjectionRowParameter;
      var expectedFullProjection = GetExpectedProjectionForNamedExpression (expectedRowParameter, "test", 0, typeof (int));
      ExpressionTreeComparer.CheckAreEqualTrees (expectedFullProjection, _commandBuilder.GetInMemoryProjectionBody());
    }

    [Test]
    public void VisitNamedExpression ()
    {
      Assert.That (_visitor.ColumnPosition, Is.EqualTo (0));

      _visitor.VisitNamedExpression (_namedExpression);

      Assert.That (_visitor.ColumnPosition, Is.EqualTo (1));

      var expectedProjection = GetExpectedProjectionForNamedExpression (_commandBuilder.InMemoryProjectionRowParameter, "test", 0, typeof (int));
      ExpressionTreeComparer.CheckAreEqualTrees (expectedProjection, _commandBuilder.GetInMemoryProjectionBody());

      Assert.That (_commandBuilder.GetCommandText(), Is.EqualTo ("@1 AS [test]"));
    }

    [Test]
    public void VisitNamedExpression_OverridesChildProjection ()
    {
      _commandBuilder.SetInMemoryProjectionBody (Expression.Constant (0));
      var nestedNamedExpression = new NamedExpression ("outer", Expression.Convert (Expression.Constant (0), typeof (double)));

      _visitor.VisitNamedExpression (nestedNamedExpression);

      Assert.That (_visitor.ColumnPosition, Is.EqualTo (1));

      var expectedProjection = GetExpectedProjectionForNamedExpression (_commandBuilder.InMemoryProjectionRowParameter, "outer", 0, typeof (double));
      ExpressionTreeComparer.CheckAreEqualTrees (expectedProjection, _commandBuilder.GetInMemoryProjectionBody ());

      Assert.That (_commandBuilder.GetCommandText (), Is.EqualTo ("@1 AS [outer]"));
    }

    [Test]
    public void VisitSqlEntityExpression ()
    {
      Assert.That (_visitor.ColumnPosition, Is.EqualTo (0));

      _visitor.VisitSqlEntityExpression (_entityExpression);

      Assert.That (_visitor.ColumnPosition, Is.EqualTo (3));

      var expectedProjection = GetExpectedProjectionForEntityExpression (_commandBuilder.InMemoryProjectionRowParameter, 0);
      ExpressionTreeComparer.CheckAreEqualTrees (expectedProjection, _commandBuilder.GetInMemoryProjectionBody ());

      Assert.That (_commandBuilder.GetCommandText (), Is.EqualTo ("[c].[ID] AS [test_ID],[c].[Name] AS [test_Name],[c].[FirstName] AS [test_FirstName]"));
    }

    [Test]
    public void VisitNewExpression_WithoutMembers ()
    {
      Assert.That (_visitor.ColumnPosition, Is.EqualTo (0));

      var newExpression = Expression.New (
          typeof (KeyValuePair<int, Cook>).GetConstructor (new[] { typeof(int), typeof (Cook)}), 
          new Expression[] { _namedExpression, _entityExpression });
      _visitor.VisitNewExpression (newExpression);

      Assert.That (_visitor.ColumnPosition, Is.EqualTo (4));

      var expectedRowParameter = _commandBuilder.InMemoryProjectionRowParameter;
      var expectedProjectionForNamedExpression = GetExpectedProjectionForNamedExpression (expectedRowParameter, "test", 0, typeof (int));
      var expectedProjectionForEntityExpression = GetExpectedProjectionForEntityExpression (expectedRowParameter, 1);
      
      var expectedProjectionForNewExpression = Expression.New (
          typeof (KeyValuePair<int, Cook>).GetConstructor (new[] { typeof (int), typeof (Cook) }),
          new Expression[] { expectedProjectionForNamedExpression, expectedProjectionForEntityExpression });

      Assert.That (
          _commandBuilder.GetCommandText (), 
          Is.EqualTo ("@1 AS [test],[c].[ID] AS [test_ID],[c].[Name] AS [test_Name],[c].[FirstName] AS [test_FirstName]"));
      ExpressionTreeComparer.CheckAreEqualTrees (expectedProjectionForNewExpression, _commandBuilder.GetInMemoryProjectionBody ());
    }

    [Test]
    public void VisitNewExpression_WithMembers ()
    {
      var keyValueType = typeof (KeyValuePair<int, Cook>);
      var newExpression = Expression.New (
          keyValueType.GetConstructor (new[] { typeof (int), typeof (Cook) }),
          new Expression[] { _namedExpression, _entityExpression },
          keyValueType.GetProperty("Key"),
          keyValueType.GetProperty("Value"));

      _visitor.VisitNewExpression (newExpression);

      var expectedRowParameter = _commandBuilder.InMemoryProjectionRowParameter;
      var expectedProjectionForNamedExpression = GetExpectedProjectionForNamedExpression (expectedRowParameter, "test", 0, typeof (int));
      var expectedProjectionForEntityExpression = GetExpectedProjectionForEntityExpression (expectedRowParameter, 1);
      
      var expectedProjectionForNewExpression = Expression.New (
          keyValueType.GetConstructor (new[] { typeof (int), typeof (Cook) }),
          new Expression[] { expectedProjectionForNamedExpression, expectedProjectionForEntityExpression },
          keyValueType.GetProperty ("Key"), keyValueType.GetProperty ("Value"));

      ExpressionTreeComparer.CheckAreEqualTrees (expectedProjectionForNewExpression, _commandBuilder.GetInMemoryProjectionBody ());
    }

    [Test]
    public void VisitConvertedBooleanExpression_ProjectionExpressonIsNull ()
    {
      var expression = new ConvertedBooleanExpression (Expression.Constant (1));
      _visitor.VisitConvertedBooleanExpression (expression);

      Assert.That (_commandBuilder.GetInMemoryProjectionBody (), Is.Null);
    }

    [Test]
    public void VisitConvertedBooleanExpression_ProjectionExpressonIsNotNull ()
    {
      var expression = new ConvertedBooleanExpression (_namedExpression);

      _visitor.VisitConvertedBooleanExpression (expression);

      var expectedProjection = Expression.Call (
          typeof (Convert).GetMethod ("ToBoolean", new[] { typeof (int) }),
          GetExpectedProjectionForNamedExpression (_commandBuilder.InMemoryProjectionRowParameter, "test", 0, typeof (int)));
      ExpressionTreeComparer.CheckAreEqualTrees (expectedProjection, _commandBuilder.GetInMemoryProjectionBody());
    }

    [Test]
    public void VisitUnaryExpression_Convert_ProjectionExpressonIsNull ()
    {
      var expression = Expression.Convert (Expression.Constant (1), typeof (double));

      _visitor.VisitUnaryExpression (expression);

      Assert.That (_commandBuilder.GetInMemoryProjectionBody (), Is.Null);
    }

    [Test]
    public void VisitUnaryExpression_Convert_ProjectionExpressonIsNotNull ()
    {
      var methodInfo = typeof (Convert).GetMethod ("ToDouble", new[] { typeof (int) });
      var expression = Expression.Convert (_namedExpression, typeof (double), methodInfo);

      _visitor.VisitUnaryExpression (expression);

      var expectedProjection = Expression.Convert (
          GetExpectedProjectionForNamedExpression (_commandBuilder.InMemoryProjectionRowParameter, "test", 0, typeof (int)),
          typeof (double),
          methodInfo);
      ExpressionTreeComparer.CheckAreEqualTrees (expectedProjection, _commandBuilder.GetInMemoryProjectionBody ());
    }

    [Test]
    public void VisitUnaryExpression_ConvertChecked_ProjectionExpressonIsNull ()
    {
      var expression = Expression.ConvertChecked (Expression.Constant (1), typeof (double));

      _visitor.VisitUnaryExpression (expression);

      Assert.That (_commandBuilder.GetInMemoryProjectionBody (), Is.Null);
    }

    [Test]
    public void VisitUnaryExpression_ConvertChecked_ProjectionExpressonIsNotNull ()
    {
      var methodInfo = typeof (Convert).GetMethod ("ToDouble", new[] { typeof (int) });
      var expression = Expression.ConvertChecked (_namedExpression, typeof (double), methodInfo);

      _visitor.VisitUnaryExpression (expression);

      var expectedProjection = Expression.ConvertChecked (
          GetExpectedProjectionForNamedExpression (_commandBuilder.InMemoryProjectionRowParameter, "test", 0, typeof (int)),
          typeof (double),
          methodInfo);
      ExpressionTreeComparer.CheckAreEqualTrees (expectedProjection, _commandBuilder.GetInMemoryProjectionBody ());
    }

    [Test]
    [ExpectedException (typeof (NotSupportedException), ExpectedMessage =
        "This SQL generator does not support queries returning groupings that result from a GroupBy operator because SQL is not suited to "
         + "efficiently return LINQ groupings. Use 'group into' and either return the items of the groupings by feeding them into an additional "
         + "from clause, or perform an aggregation on the groupings.", MatchType = MessageMatch.Contains)]
    public void VisitSqlGroupingSelectExpression ()
    {
      var expression = SqlStatementModelObjectMother.CreateSqlGroupingSelectExpression ();
      _visitor.VisitSqlGroupingSelectExpression (expression);
    }

    private MethodCallExpression GetExpectedProjectionForEntityExpression (ParameterExpression expectedRowParameter, int columnPositionStart)
    {
      return Expression.Call (
          expectedRowParameter,
          typeof (IDatabaseResultRow).GetMethod ("GetEntity").MakeGenericMethod (typeof (Cook)),
          Expression.Constant (new[] { 
              new ColumnID ("test_ID", columnPositionStart), 
              new ColumnID ("test_Name", columnPositionStart + 1), 
              new ColumnID ("test_FirstName", columnPositionStart + 2) }));
    }

    private MethodCallExpression GetExpectedProjectionForNamedExpression (
        ParameterExpression rowParameter,
        string name,
        int columnPosoitionStart,
        Type expressionType)
    {
      return Expression.Call (
          rowParameter,
          typeof (IDatabaseResultRow).GetMethod ("GetValue").MakeGenericMethod (expressionType),
          Expression.Constant (new ColumnID (name ?? "value", columnPosoitionStart)));
    }
  }
}